# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import ast
import json
import os
import subprocess
from pathlib import Path
from typing import Callable, List, Tuple, Union
from kqcircuits.defaults import STARTUPINFO, XSECTION_PROCESS_PATH
from kqcircuits.pya_resolver import pya, klayout_executable_command
from kqcircuits.simulations.export.util import export_layers
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder


def xsection_call(
    input_oas: Path,
    output_oas: Path,
    cut1: pya.DPoint,
    cut2: pya.DPoint,
    process_path: Path = XSECTION_PROCESS_PATH,
    parameters_path: Path = None,
) -> None:
    """Calls on KLayout to run the XSection plugin

    Args:
        input_oas: Input OAS file (top-down geometry)
        output_oas: Output OAS file (Cross-section of input geometry)
        cut1: DPoint of first endpoint of the cross-section cut
        cut2: DPoint of second endpoint of the cross-section cut
        process_path: XSection process file that defines cross-section etching depths etc
        parameters_path: If process_path points to kqc_process.xs,
            parameters_path should point to the XSection parameters json file
            containing sweeped parameters and layer information.
    """
    if os.name == "nt":
        klayout_dir_name = "KLayout"
    elif os.name == "posix":
        klayout_dir_name = ".klayout"
    else:
        raise SystemError("Error: unsupported operating system")
    xsection_plugin_path = os.path.join(os.path.expanduser("~"), klayout_dir_name, "salt/xsection/macros/xsection.lym")
    cut_string = f"{cut1.x},{cut1.y};{cut2.x},{cut2.y}"

    if not klayout_executable_command():
        raise Exception("Can't find klayout executable command!")
    if not Path(xsection_plugin_path).is_file():
        raise Exception("The 'xsection' plugin is missing in KLayout! Go to 'Tools->Manage Packages' to install it.")

    # Hack: Weird prefix keeps getting added when path is converted to string which breaks the ruby plugin
    xs_run = str(process_path).replace("\\\\?\\", "")
    xs_params = str(parameters_path).replace("\\\\?\\", "")
    # When debugging, remove '-z' argument to see ruby error messages
    subprocess.run(
        [
            klayout_executable_command(),
            input_oas.absolute(),
            "-z",
            "-nc",
            "-rx",
            "-r",
            xsection_plugin_path,
            "-rd",
            f"xs_run={xs_run}",
            "-rd",
            f"xs_params={xs_params}",
            "-rd",
            f"xs_cut={cut_string}",
            "-rd",
            f"xs_out={output_oas.absolute()}",
        ],
        check=True,
        startupinfo=STARTUPINFO,
    )


def _oxidise_layers(simulation, ma_thickness, ms_thickness, sa_thickness):
    """Take the cross section geometry and add oxide layers between substrate, metal and vacuum.
    Will add geometry around metals and etch away substrate to insert oxide geometry.
    """
    substrate_layers = [layer for layer in simulation.layout.layer_infos() if layer.name.endswith("_substrate")]
    substrate = _combine_region_from_layers(simulation, substrate_layers)
    metal_layers = [
        layer
        for layer in simulation.layout.layer_infos()
        if layer.name in ["b_ground", "t_ground", "b_signal", "t_signal"]
    ]
    metal_layers += [layer for layer in simulation.layout.layer_infos() if layer.name.startswith("b_signal_")]
    metal_layers += [layer for layer in simulation.layout.layer_infos() if layer.name.startswith("t_signal_")]
    metals = _combine_region_from_layers(simulation, metal_layers)
    metal_edges = metals.edges()
    substrate_edges = substrate.edges()

    ma_edges = []
    for metal_edge in metal_edges:
        if not _edge_on_the_box_border(metal_edge.to_dtype(simulation.layout.dbu), simulation.box):
            ma_edges.extend(_remove_shared_points(metal_edge, substrate_edges, True))

    sa_edges, ms_edges = [], []
    for substrate_edge in substrate_edges:
        if not _edge_on_the_box_border(substrate_edge.to_dtype(simulation.layout.dbu), simulation.box):
            sa_edges.extend(_remove_shared_points(substrate_edge, metal_edges, True))
            ms_edges.extend(_remove_shared_points(substrate_edge, sa_edges, False))

    ma_layer = _thicken_edges(simulation, ma_edges, ma_thickness, True)
    ms_layer = _thicken_edges(simulation, ms_edges, ms_thickness, False)
    sa_layer = _thicken_edges(simulation, sa_edges, sa_thickness, True)
    sa_layer -= ma_layer  # MA layer takes precedence over SA layer

    # Etch and replace substrate layer regions
    if ms_thickness > 0.0 or sa_thickness > 0.0:
        for substrate_layer in substrate_layers:
            substrate_region = pya.Region(simulation.cell.shapes(simulation.layout.layer(substrate_layer)))
            simulation.cell.shapes(simulation.layout.layer(substrate_layer)).clear()
            simulation.cell.shapes(simulation.layout.layer(substrate_layer)).insert(substrate_region - ms_layer)

    if ma_thickness > 0.0:
        simulation.cell.shapes(simulation.get_sim_layer("ma_layer")).insert(ma_layer)
    if ms_thickness > 0.0:
        simulation.cell.shapes(simulation.get_sim_layer("ms_layer")).insert(ms_layer)
    if sa_thickness > 0.0:
        simulation.cell.shapes(simulation.get_sim_layer("sa_layer")).insert(sa_layer)


def create_xsections_from_simulations(
    simulations: List[Simulation],
    output_path: Path,
    cuts: Union[Tuple[pya.DPoint, pya.DPoint], List[Tuple[pya.DPoint, pya.DPoint]]],
    process_path: Path = XSECTION_PROCESS_PATH,
    post_processing_function: Callable[[CrossSectionSimulation], None] = None,
    oxidise_layers_function: Callable[[CrossSectionSimulation, float, float, float], None] = _oxidise_layers,
    ma_permittivity: float = 0,
    ms_permittivity: float = 0,
    sa_permittivity: float = 0,
    ma_thickness: float = 0,
    ms_thickness: float = 0,
    sa_thickness: float = 0,
    vertical_cull: Union[None, Tuple[float, float]] = None,
    mer_box: Union[None, pya.DBox] = None,
    london_penetration_depth: float = 0,
    magnification_order: int = 0,
) -> List[Simulation]:
    """Create cross-sections of all simulation geometries in the list.
    Will set 'box' and 'cell' parameters according to the produced cross-section geometry data.

    Args:
        simulations: List of Simulation objects, usually produced by a sweep
        output_path: Path for the exported simulation files
        cuts: 1. A tuple (p1, p2), where p1 and p2 are endpoints of a cross-section cut or
              2. a list of such tuples such that each Simulation object gets an individual cut
        process_path: XSection process file that defines cross-section etching depths etc
        post_processing_function: Additional function to post-process the cross-section geometry.
            Defaults to None, in which case no post-processing is performed.
            The function takes a CrossSectionSimulation object as argument
        oxidise_layers_function: Set this argument if you have a custom way of introducing
            oxidization layers to the cross-section metal deposits and substrate.
            See expected function signature from pyhints
        ma_permittivity: Permittivity of metal–vacuum (air) interface
        ms_permittivity: Permittivity of metal–substrate interface
        sa_permittivity: Permittivity of substrate–vacuum (air) interface
        ma_thickness: Thickness of metal–vacuum (air) interface
        ms_thickness: Thickness of metal–substrate interface
        sa_thickness: Thickness of substrate–vacuum (air) interface
        vertical_cull: Tuple of two y-coordinates, will cull all geometry not in-between the y-coordinates.
            None by default, which means all geometry is retained.
        mer_box: If set as some pya.DBox, will create a specified box as metal edge region,
            meaning that the geometry inside the region are separated into different layers with '_mer' suffix
        london_penetration_depth: London penetration depth of the superconducting material
        magnification_order: Increase magnification of simulation geometry to accomodate more precise spacial units.
            0 =   no magnification with 1e-3 dbu
            1 =  10x magnification with 1e-4 dbu
            2 = 100x magnification with 1e-5 dbu etc
            Consider setting non-zero value when using oxide layers with < 1e-3 layer thickness or
            taking cross-sections of thin objects

    Returns:
        List of CrossSectionSimulation objects for each Simulation object in simulations
    """
    if isinstance(cuts, Tuple):
        cuts = [cuts] * len(simulations)
    if len(simulations) != len(cuts):
        raise Exception("Number of cuts did not match the number of simulations")
    if any(len(simulation.get_parameters()["face_stack"]) not in (1, 2) for simulation in simulations):
        raise Exception("Only single face and flip chip cross section simulations currently supported")

    xsection_dir = output_path.joinpath("xsection_tmp")
    xsection_dir.mkdir(parents=True, exist_ok=True)

    layout = pya.Layout()
    load_opts = _load_layout_options_for_xsection_output()
    for simulation, cut in zip(simulations, cuts):
        xsection_parameters = _dump_xsection_parameters(xsection_dir, simulation)
        simulation_file = xsection_dir / f"original_{simulation.cell.name}.oas"
        xsection_file = xsection_dir / f"xsection_{simulation.cell.name}.oas"
        export_layers(str(simulation_file), simulation.layout, [simulation.cell], output_format="OASIS", layers=None)
        xsection_call(simulation_file, xsection_file, cut[0], cut[1], process_path, xsection_parameters)

        layout.read(str(xsection_file), load_opts)
        xsection_cell = layout.top_cells()[-1]
        xsection_cell.name = simulation.cell.name

    _clean_tmp_xsection_directory(xsection_dir, simulations)
    # Collect cross section simulation sweeps
    return [
        _construct_cross_section_simulation(
            layout,
            xsection_cell,
            simulations[idx],
            post_processing_function,
            oxidise_layers_function,
            ma_permittivity,
            ms_permittivity,
            sa_permittivity,
            ma_thickness,
            ms_thickness,
            sa_thickness,
            vertical_cull,
            mer_box,
            london_penetration_depth,
            magnification_order,
        )
        for idx, xsection_cell in enumerate(layout.top_cells())
    ]


def separate_signal_layer_shapes(simulation: Simulation, sort_key: Callable[[pya.Shape], float] = None):
    """Separate shapes in signal layer to their own dedicated signal layers for each face

    Args:
        simulation: A Simulation object where the layer will be separated
        sort_key: A function that, given a Shape object, returns a number.
            Shapes are sorted according to the number in increasing order.
            If None, picks a point in shape polygon, sorts points top to bottom then tie-breaks left to right
    """
    if sort_key is None:

        def sort_key(shape):
            point_in_shape = list(shape.polygon.each_point_hull())[0]
            return (-point_in_shape.y, point_in_shape.x)

    signal_index = 1
    gen_free_layer_slots = free_layer_slots(simulation.layout)
    for face in simulation.face_ids:
        signal_layer = find_layer_by_name(f"{face}_signal", simulation.layout)
        if signal_layer is None:
            continue
        signal_layer_idx = simulation.layout.layer(signal_layer)
        for shape in sorted(simulation.cell.each_shape(signal_layer_idx), key=sort_key):
            # Reuse layer if it already used in layout
            signal_layer = find_layer_by_name(f"{face}_signal_{signal_index}", simulation.layout)
            # If no such layer, find next available layer index
            if signal_layer is None:
                layer_index = next(gen_free_layer_slots)
                signal_layer = pya.LayerInfo(layer_index, 0, f"{face}_signal_{signal_index}")
            simulation.cell.shapes(simulation.layout.layer(signal_layer)).insert(shape)
            signal_index += 1
        simulation.cell.clear(signal_layer_idx)


def find_layer_by_name(layer_name, layout):
    """Returns layerinfo if there already is a layer by layer_name in layout. None if no such layer exists"""
    for l in layout.layer_infos():
        if l.datatype == 0 and layer_name == l.name:
            return l
    return None


def free_layer_slots(layout):
    """A generator of available layer slots"""
    layer_index = 0
    reserved_layer_ids = [l.layer for l in layout.layer_infos() if l.datatype == 0]
    while True:
        layer_index += 1
        if layer_index in reserved_layer_ids:
            continue
        yield layer_index


def visualise_xsection_cut_on_original_layout(
    simulations: List[Simulation],
    cuts: Union[Tuple[pya.DPoint, pya.DPoint], List[Tuple[pya.DPoint, pya.DPoint]]],
    cut_label: str = "cut",
    width_ratio: float = 0.0,
):
    """Visualise requested xsection cuts on the original simulation layout.

    Will add a rectangle between two points of the cut, and two text points into layer "xsection_cut"::

        * f"{cut_label}_1" representing the left side of the cross section simulation
        * f"{cut_label}_2" representing the right side of the cross section simulation

    In case the export takes xsections for one simulation multiple times, this function
    can be called on same simulation sweep multiple times so that multiple cuts can be visualised
    in the same layout. In such case it is recommended to differentiate the cuts using `cut_label`.

    Args:
        simulations: list of simulations from which xsections are taken. After this call these simulations
            will be modified to include the visualised cuts.
        cuts: 1. A tuple (p1, p2), where p1 and p2 are endpoints of a cross-section cut or
              2. a list of such tuples such that each Simulation object gets an individual cut
        cut_label: prefix of the two text points shown for the cut
        width_ratio: rectangles visualising cuts will have a width of length of the cut multiplied by width_ratio
    """
    if isinstance(cuts, Tuple):
        cuts = [cuts] * len(simulations)
    if len(simulations) != len(cuts):
        raise Exception("Number of cuts did not match the number of simulations")
    for simulation, cut in zip(simulations, cuts):
        cut_length = (cut[1] - cut[0]).length()
        marker = pya.Region(pya.DPath(cut, cut_length * width_ratio).to_itype(simulation.layout.dbu))
        # Not using get_sim_layer since this geometry is not part of any simulation
        cut_visualisation_layer = simulation.layout.layer("xsection_cut")
        simulation.cell.shapes(cut_visualisation_layer).insert(marker)
        simulation.cell.shapes(cut_visualisation_layer).insert(pya.DText(f"{cut_label}_1", cut[0].x, cut[0].y))
        simulation.cell.shapes(cut_visualisation_layer).insert(pya.DText(f"{cut_label}_2", cut[1].x, cut[1].y))


def _load_layout_options_for_xsection_output():
    load_opts = pya.LoadLayoutOptions()
    load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
    return load_opts


def _remap_face(layer_name, faces_of_flipchip):
    """Rename face to b_*, t_* convention based on faces_of_flipchip list"""
    if len(faces_of_flipchip) > 0 and layer_name.startswith(f"{faces_of_flipchip[0]}_"):
        return f"b_{layer_name[len(faces_of_flipchip[0]) + 1:]}"
    if len(faces_of_flipchip) > 1 and layer_name.startswith(f"{faces_of_flipchip[1]}_"):
        return f"t_{layer_name[len(faces_of_flipchip[1]) + 1:]}"
    return layer_name


def _dump_xsection_parameters(xsection_dir, simulation):
    """If we're sweeping xsection specific parameters,
    dump them in external file for xsection process file to pick up
    """
    simulation_params = {
        param_name: param_value
        for param_name, param_value in simulation.get_parameters().items()
        if not isinstance(param_value, pya.DBox)
    }  # Hack: ignore non-serializable params
    # Also dump all used layers in the simulation cell
    sim_layers = {
        _remap_face(l.name, simulation_params["face_stack"]): f"{l.layer}/{l.datatype}"
        for l in simulation.layout.layer_infos()
    }
    # Find avaiable layer numbers for substrate layers
    gen_free_layer_slots = free_layer_slots(simulation.layout)
    sim_layers["b_substrate"] = f"{next(gen_free_layer_slots)}/0"
    sim_layers["t_substrate"] = f"{next(gen_free_layer_slots)}/0"
    simulation_params["sim_layers"] = sim_layers
    xsection_parameters_file = xsection_dir / f"parameters_{simulation.cell.name}.json"
    with open(xsection_parameters_file, "w") as sweep_file:
        json.dump(simulation_params, sweep_file, cls=GeometryJsonEncoder)
    return xsection_parameters_file


def _clean_tmp_xsection_directory(xsection_dir, simulations):
    for simulation in simulations:
        if os.path.exists(xsection_dir / f"original_{simulation.cell.name}.oas"):
            os.remove(xsection_dir / f"original_{simulation.cell.name}.oas")
        if os.path.exists(xsection_dir / f"xsection_{simulation.cell.name}.oas"):
            os.remove(xsection_dir / f"xsection_{simulation.cell.name}.oas")
        if os.path.exists(xsection_dir / f"parameters_{simulation.cell.name}.json"):
            os.remove(xsection_dir / f"parameters_{simulation.cell.name}.json")
    if os.path.exists(xsection_dir):
        os.rmdir(xsection_dir)


def _combine_region_from_layers(simulation, layers):
    """Produce a region combined from regions in layers list"""
    region = pya.Region()
    for layer in layers:
        region += pya.Region(simulation.cell.shapes(simulation.layout.layer(layer)))
    return region


def _edge_on_the_box_border(edge, box):
    """True if edge is exactly at the rim of the box. edge must be of class pya.DEdge"""
    return (
        (edge.x1 == box.p1.x and edge.x2 == box.p1.x)
        or (edge.x1 == box.p2.x and edge.x2 == box.p2.x)
        or (edge.y1 == box.p1.y and edge.y2 == box.p1.y)
        or (edge.y1 == box.p2.y and edge.y2 == box.p2.y)
    )


def _cut_edge(target_edge, source_edge, extra_edges):
    """Cut an end of the target_edge with source_edge.

    If source_edge leaves behind two ends of the target_edge,
    the second edge bit is stored in extra_edges.

    Each edge should be in integer form (pya.Edge)
    """
    # Copy target_edge to not modify the original edge instance
    result_edge = pya.Edge(target_edge.p1.x, target_edge.p1.y, target_edge.p2.x, target_edge.p2.y)
    if result_edge.contains_excl(source_edge.p1):
        if result_edge.contains_excl(source_edge.p2) and source_edge.p2 != result_edge.p2:
            extra_edges.append(pya.Edge(source_edge.p2, result_edge.p2))
        result_edge.p2 = source_edge.p1
    elif result_edge.contains_excl(source_edge.p2):
        result_edge.p1 = source_edge.p2
    return result_edge


def _remove_shared_points(target_edge, acting_edges, is_adjacent):
    """Remove all points shared by target_edge and edges in acting_edges

    Returns a set of continuous edges that are not contained by acting_edges.
    Set is_adjacent to True if the shape of acting_edges is adjacent to the shape
    from which target_edge was taken. Set to False if the shapes are on top of eah other.

    Each edge should be in integer form (pya.Edge)
    """
    edge_bits = [target_edge]
    for acting_edge in acting_edges:
        # Set acting_edge to point to same direction as target_edge
        if is_adjacent:
            acting_edge = acting_edge.swapped_points()
        # Consider edges if they share points, which means they are parallel and have same displacement
        if acting_edge.is_parallel(target_edge):
            # Remove edge bits if they are completely covered by acting_edge
            edge_bits = [e for e in edge_bits if not (acting_edge.contains(e.p1) and acting_edge.contains(e.p2))]
            extra_edge_bits = []  # Collect extra edge bits here
            edge_bits = [_cut_edge(e, acting_edge, extra_edge_bits) for e in edge_bits]
            edge_bits.extend(extra_edge_bits)  # Add extra bits
            edge_bits = [e for e in edge_bits if e.p1 != e.p2]  # Remove zero length edge bits
    return edge_bits


def _normal_of_edge(simulation, p1, p2, scale):
    """Returns a normal of edge p1->p2.

    If (p1, p2) are in same polygon in given order,
    the returned normal will stick out of polygon.
    The magnitude of the normal will be set to `scale`.

    p1, p2 are pya.Point (integer) objects, however
    scale is scaled according to pya.DPoint domain.
    """
    edge_dir = p2 - p1
    normal = pya.Point(-edge_dir.y, edge_dir.x)
    dnormal = normal.to_dtype(simulation.layout.dbu)
    return (dnormal * (scale / dnormal.abs())).to_itype(simulation.layout.dbu)


def _thicken_edges(simulation, edges, thickness, grow):
    """Take edges and add thickness to produce a region.

    Set grow to True to grow the region outward, False to grow inward
    Each edge should be in integer form (pya.Edge)
    """
    if thickness <= 0.0:  # Don't do anything if no thickness
        return pya.Region()
    # Construct a graph from the edges to find paths
    # Start by finding start points for paths
    start_points = [e.p1 for e in edges if e.p1 not in [e2.p2 for e2 in edges]]
    path_graph = {}
    for edge in edges:
        path_graph[edge.p1] = edge

    result_region = pya.Region()
    processed_edges = []
    # Take each start_point and follow the path until the end
    for current_point in start_points:
        inner_path = [current_point]
        normals = []
        while True:
            current_edge = path_graph[current_point]
            processed_edges.append(current_edge)
            # First collect path points for the region polygon
            inner_path.append(current_edge.p2)
            # Also collect their normals
            normals.append(
                (1.0 if grow else -1.0) * _normal_of_edge(simulation, current_edge.p1, current_edge.p2, thickness)
            )
            # At the end point, terminate
            if current_edge.p2 not in path_graph:
                break
            # Otherwise proceed to next point in path
            current_point = current_edge.p2
        # Connect to the second layer of the path to add thickness
        outer_path = [inner_path[-1] + normals[-1]]
        # Backtrack the path for the second layer of the polygon
        for idx in range(len(normals) - 1, 0, -1):
            normal_sum = normals[idx] + normals[idx - 1]  # Sum normals of surrounding edges of the point
            outer_path.append(inner_path[idx] + normal_sum)
        outer_path.append(inner_path[0] + normals[0])
        result_region += pya.Region(pya.Polygon(inner_path + outer_path))

    # Handle edges in loops separately
    loop_edges = [e for e in edges if e not in processed_edges]
    processed_edges = []
    for start_edge in loop_edges:
        if start_edge in processed_edges:
            continue
        current_edge = start_edge
        loop = [current_edge.p1]
        while current_edge.p2 != start_edge.p1:
            loop.append(current_edge.p2)
            current_edge = path_graph[current_edge.p2]
            processed_edges.append(current_edge)
        loop_poly = pya.Polygon(loop)
        if grow:
            # We are growing on the rim of the loop. Take the rim, then copy the
            # shrinked rim, then subtract shrinked rim from original rim.
            # To shrink, we have to rely on normals
            loop_sized = []
            for i, p in enumerate(loop):
                j = 0 if i + 1 == len(loop) else i + 1
                normal_next = _normal_of_edge(simulation, p, loop[j], thickness)
                normal_prev = _normal_of_edge(simulation, loop[i - 1], p, thickness)
                loop_sized.append(p + normal_prev + normal_next)
            loop_sized = pya.Polygon(loop_sized)
        else:
            # We are "growing" out of the rim of the loop. Take the rim,
            # then copy the magnified rim, subtracting original rim from
            # the magnified rim. We can just use `sized` method.
            loop_sized = loop_poly.to_dtype(simulation.layout.dbu)
            loop_sized = loop_sized.sized(thickness)
            loop_sized = loop_sized.to_itype(simulation.layout.dbu)
        if grow:
            result_region += pya.Region(loop_poly) - pya.Region(loop_sized)
        else:
            result_region += pya.Region(loop_sized) - pya.Region(loop_poly)
    return result_region


def _iterate_layers_and_modify_region(xsection_cell, process_region):
    """Iterates over all (non-empty) layers in xsection_cell
    and replaces the region in that layer with process_region(region, layer)
    """
    for layer in xsection_cell.layout().layer_infos():
        region = pya.Region(xsection_cell.shapes(xsection_cell.layout().layer(layer)))
        if region.is_empty():
            continue
        xsection_cell.shapes(xsection_cell.layout().layer(layer)).clear()
        xsection_cell.shapes(xsection_cell.layout().layer(layer)).insert(process_region(region, layer))


def _construct_cross_section_simulation(
    layout,
    xsection_cell,
    simulation,
    post_processing_function,
    oxidise_layers_function,
    ma_permittivity,
    ms_permittivity,
    sa_permittivity,
    ma_thickness,
    ms_thickness,
    sa_thickness,
    vertical_cull,
    mer_box,
    london_penetration_depth,
    magnification_order,
):
    """Produce CrossSectionSimulation object"""
    if magnification_order > 0:
        layout.dbu = 10 ** (-3 - magnification_order)
        xsection_cell.transform(pya.DCplxTrans(10**magnification_order))
    xsection_parameters = simulation.get_parameters()
    xsection_parameters["london_penetration_depth"] = london_penetration_depth
    cell_bbox = xsection_cell.dbbox()
    # Disabled for single face and flip-chip cases
    # cell_bbox.p1 -= pya.DPoint(0, xsection_parameters['lower_box_height'])
    if len(xsection_parameters["face_stack"]) == 1:
        cell_bbox.p2 += pya.DPoint(0, xsection_parameters["upper_box_height"])
    if vertical_cull is not None:
        cell_bbox.p1 = pya.DPoint(cell_bbox.p1.x, min(vertical_cull))
        cell_bbox.p2 = pya.DPoint(cell_bbox.p2.x, max(vertical_cull))
    xsection_parameters["box"] = cell_bbox
    xsection_parameters["cell"] = xsection_cell
    xsection_simulation = CrossSectionSimulation(layout, **xsection_parameters)
    # Keep all parameters given in simulations for JSON
    for k, v in xsection_parameters.items():
        setattr(xsection_simulation, k, v)
    xsection_simulation.xsection_source_class = type(simulation)
    xsection_simulation.register_cell_layers_as_sim_layers()

    material_dict = xsection_parameters["material_dict"]
    material_dict = ast.literal_eval(material_dict) if isinstance(material_dict, str) else material_dict
    substrate_material = xsection_parameters["substrate_material"]
    b_substrate_permittivity = material_dict[substrate_material[0]]["permittivity"]
    xsection_simulation.set_permittivity("b_substrate", b_substrate_permittivity)
    if len(xsection_parameters["face_stack"]) == 2:
        t_substrate_permittivity = b_substrate_permittivity
        if len(substrate_material) > 1:
            t_substrate_permittivity = material_dict[substrate_material[1]]["permittivity"]
        xsection_simulation.set_permittivity("t_substrate", t_substrate_permittivity)

    if post_processing_function is not None:
        post_processing_function(xsection_simulation)

    oxidise_layers_function(xsection_simulation, ma_thickness, ms_thickness, sa_thickness)

    if vertical_cull is not None:

        def _cull_region_vertically(region, layer):  # pylint: disable=unused-argument
            return region & cell_bbox.to_itype(xsection_cell.layout().dbu)

        _iterate_layers_and_modify_region(xsection_cell, _cull_region_vertically)

    if mer_box is not None:
        regions_to_update = {}

        def _separate_region_in_mer_box(region, layer):
            region_in_box = region & mer_box.to_itype(xsection_cell.layout().dbu)
            regions_to_update[f"{layer.name}_mer"] = region_in_box
            return region - mer_box.to_itype(xsection_cell.layout().dbu)

        _iterate_layers_and_modify_region(xsection_cell, _separate_region_in_mer_box)
        vacuum_in_box = pya.Region(mer_box.to_itype(xsection_cell.layout().dbu))
        for layer, region in regions_to_update.items():
            vacuum_in_box -= region
            xsection_cell.shapes(xsection_simulation.get_sim_layer(layer)).insert(region)
        xsection_cell.shapes(xsection_simulation.get_sim_layer("vacuum_mer")).insert(vacuum_in_box)

    if ma_thickness > 0.0:
        xsection_simulation.set_permittivity("ma_layer", ma_permittivity)
    if ms_thickness > 0.0:
        xsection_simulation.set_permittivity("ms_layer", ms_permittivity)
    if sa_thickness > 0.0:
        xsection_simulation.set_permittivity("sa_layer", sa_permittivity)
    return xsection_simulation



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from dataclasses import dataclass
from typing import Optional, Union, List
from kqcircuits.simulations.export.solution import Solution


@dataclass
class AnsysSolution(Solution):
    """
    A Base class for Ansys Solution and export parameters

    Args:
        ansys_tool: Determines whether to use 'hfss' (s-parameters), 'q3d', 'current', 'voltage', or 'eigenmode'.
        frequency_units: Units of frequency.
        frequency: Frequency for mesh refinement. To set up multifrequency analysis in HFSS use list of numbers.
        max_delta_s: Stopping criterion in HFSS simulation.
        max_delta_e: Stopping criterion in current or voltage excitation simulation.
        percent_error: Stopping criterion in Q3D simulation.
        percent_refinement: Percentage of mesh refinement on each iteration.
        maximum_passes: Maximum number of iterations in simulation.
        minimum_passes: Minimum number of iterations in simulation.
        minimum_converged_passes: Determines how many iterations have to meet the stopping criterion to stop simulation.
        sweep_enabled: Determines if HFSS frequency sweep is enabled.
        sweep_start: The lowest frequency in the sweep.
        sweep_end: The highest frequency in the sweep.
        sweep_count: Number of frequencies in the sweep.
        sweep_type: choices are "interpolating", "discrete" or "fast"
        max_delta_f: Maximum allowed relative difference in eigenfrequency (%). Used when ``ansys_tool`` is *eigenmode*.
        n_modes: Number of eigenmodes to solve. Used when ``ansys_tool`` is 'eigenmode'.
        mesh_size: Dictionary to determine manual mesh refinement on layers. Set key as the layer name and value as the
            maximal mesh element length inside the layer.
        simulation_flags: Optional export processing, given as list of strings. See Simulation Export in docs.
        ansys_project_template: path to the simulation template
        integrate_energies: Calculate energy integrals over each layer and save them into a file
        integrate_magnetic_flux: Integrate magnetic fluxes through each non-pec sheet and save them into a file
        hfss_capacitance_export: If True, the capacitance matrices are exported from HFSS simulations
    """

    ansys_tool: str = "hfss"  # to be ClassVar
    frequency_units: str = "GHz"
    frequency: Union[float, List[float]] = 5
    max_delta_s: float = 0.1
    max_delta_e: float = 0.1
    percent_error: float = 1
    percent_refinement: float = 30.0
    maximum_passes: int = 12
    minimum_passes: int = 1
    minimum_converged_passes: int = 1
    sweep_enabled: bool = True
    sweep_start: float = 0
    sweep_end: float = 10
    sweep_count: int = 101
    sweep_type: str = "interpolating"
    max_delta_f: float = 0.1
    n_modes: int = 2
    mesh_size: Optional[dict] = None
    simulation_flags: Optional[List[str]] = None
    ansys_project_template: Optional[str] = None
    integrate_energies: bool = False
    integrate_magnetic_flux: bool = False
    hfss_capacitance_export: bool = False

    def get_solution_data(self):
        """Return the solution data in dictionary form."""
        return {
            "solution_name": self.name,
            "ansys_tool": self.ansys_tool,
            "analysis_setup": {
                "frequency_units": self.frequency_units,
                "frequency": self.frequency,
                "max_delta_s": self.max_delta_s,  # stopping criterion for HFSS
                "max_delta_e": self.max_delta_e,  # stopping criterion for current or voltage excitation simulation
                "percent_error": self.percent_error,  # stopping criterion for Q3D
                "percent_refinement": self.percent_refinement,
                "maximum_passes": self.maximum_passes,
                "minimum_passes": self.minimum_passes,
                "minimum_converged_passes": self.minimum_converged_passes,
                "sweep_enabled": self.sweep_enabled,
                "sweep_start": self.sweep_start,
                "sweep_end": self.sweep_end,
                "sweep_count": self.sweep_count,
                "sweep_type": self.sweep_type,
                "max_delta_f": self.max_delta_f,
                "n_modes": self.n_modes,
            },
            "mesh_size": {} if self.mesh_size is None else self.mesh_size,
            "simulation_flags": [] if self.simulation_flags is None else self.simulation_flags,
            "integrate_energies": self.integrate_energies,
            "integrate_magnetic_flux": self.integrate_magnetic_flux,
            "hfss_capacitance_export": self.hfss_capacitance_export,
            **({} if self.ansys_project_template is None else {"ansys_project_template": self.ansys_project_template}),
        }



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import os
import stat
import logging
import json
import argparse

from pathlib import Path
from typing import Sequence, Union, Tuple, Dict, Optional

from kqcircuits.simulations.export.simulation_export import (
    copy_content_into_directory,
    get_post_process_command_lines,
    get_combined_parameters,
)
from kqcircuits.simulations.export.util import export_layers
from kqcircuits.util.export_helper import write_commit_reference_file
from kqcircuits.defaults import ELMER_SCRIPT_PATHS, KQC_REMOTE_ACCOUNT
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder
from kqcircuits.simulations.export.elmer.elmer_solution import ElmerSolution
from kqcircuits.simulations.post_process import PostProcess


def export_elmer_json(
    simulation: Union[Simulation, CrossSectionSimulation], solution: ElmerSolution, path: Path, workflow: dict
):
    """
    Export Elmer simulation into json and gds files.

    Args:
        simulation: The simulation to be exported.
        solution: The solution to be exported.
        path: Location where to write json and gds files.
        workflow: Parameters for simulation workflow

    Returns:
         Path to exported json file.
    """
    is_cross_section = isinstance(simulation, CrossSectionSimulation)

    if simulation is None or not isinstance(simulation, (Simulation, CrossSectionSimulation)):
        raise ValueError("Cannot export without simulation")

    # collect data for .json file
    if is_cross_section:
        layers = simulation.layer_dict

    # write .gds file
    gds_file = simulation.name + ".gds"
    gds_file_path = str(path.joinpath(gds_file))
    export_layers(
        gds_file_path,
        simulation.layout,
        [simulation.cell],
        output_format="GDS2",
        layers=layers.values() if is_cross_section else simulation.get_layers(),
    )

    sim_data = simulation.get_simulation_data()
    sol_data = solution.get_solution_data()
    full_name = simulation.name + solution.name

    if is_cross_section:
        if sim_data.get("london_penetration_depth", 0.0) > 0:
            sif_names = [f"{full_name}_C", f"{full_name}_L"]
        else:
            sif_names = [f"{full_name}_C", f"{full_name}_C0"]
    elif solution.tool == "wave_equation":
        if solution.sweep_type == "interpolating":
            sif_names = []
        else:
            sif_names = [full_name + "_f" + str(f).replace(".", "_") for f in sol_data["frequency"]]
    else:
        sif_names = [full_name]

    json_data = {
        "name": full_name,
        "workflow": workflow,
        **sim_data,
        **sol_data,
        **({"layers": {k: (v.layer, v.datatype) for k, v in layers.items()}} if is_cross_section else {}),
        "sif_names": sif_names,
        "gds_file": gds_file,
        "parameters": get_combined_parameters(simulation, solution),
    }

    # write .json file
    json_file_path = str(path.joinpath(full_name + ".json"))
    with open(json_file_path, "w") as fp:
        json.dump(json_data, fp, cls=GeometryJsonEncoder, indent=4)

    return json_file_path


def export_elmer_script(
    json_filenames,
    path: Path,
    workflow=None,
    file_prefix="simulation",
    execution_script="scripts/run.py",
    post_process=None,
):
    """
    Create script files for running one or more simulations.
    Create also a main script to launch all the simulations at once.

    Args:
        json_filenames: List of paths to json files to be included into the script.
        path: Location where to write the script file.
        workflow: Parameters for simulation workflow
        file_prefix: Name of the script file to be created.
        execution_script: The script file to be executed.
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations

    Returns:

        Path of exported main script file
    """

    if workflow is None:
        workflow = dict()
    sbatch = "sbatch_parameters" in workflow

    python_executable = workflow.get("python_executable", "python")
    main_script_filename = str(path.joinpath(file_prefix + ".sh"))

    path.joinpath("log_files").mkdir(parents=True, exist_ok=True)

    if sbatch:

        def _multiply_time(time_str, multiplier):
            """
            Helper function to multiply a time of format "HH:MM:SS" with a constant. In
            this case, we multiply timeout per simulation by the number of simulations

            Args:
                time_str (str): Time in format "HH:MM:SS"
                multiplier (float): multiplier

            Returns:
                New time string in format "HH:MM:SS"
            """
            time_str = time_str.strip()
            if len(time_str) != 8:
                raise ValueError('Invalid sbatch/slurm time formatting! Format has to be "HH:MM:SS"')
            hours = int(int(time_str[0:2]) * multiplier)
            minutes = int(int(time_str[3:5]) * multiplier)
            seconds = int(int(time_str[6:8]) * multiplier)
            if seconds > 60:
                minutes = minutes + seconds // 60
                seconds = seconds % 60
            if minutes > 60:
                hours = hours + minutes // 60
                minutes = minutes % 60
            return "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)

        def _multiply_mem(mem_str, multiplier):
            """
            Helper function to multiply a memory specification with a constant. If multiplier < 1,
            will convert to using smaller units if possible

            Args:
                mem_str (str): Amount of memory in format "1234X" where "X" specifies the unit
                               and anything before is a number specifying the amount
                multiplier (float): multiplier

            Returns:
                New memory string
            """
            original_memstr = mem_str.strip()
            unit = original_memstr[-1]
            if unit in ("K", "M", "G", "T"):
                original_mem_int = int(original_memstr.partition(unit)[0])
            else:
                original_mem_int = int(original_memstr)
                unit = "M"
            downconversion = {"M": "K", "G": "M", "T": "G"}
            if multiplier < 1.0 and unit != "K":
                elmer_mem_per_f = int(original_mem_int * 1024 * multiplier)
                elmer_mem_per_f = str(elmer_mem_per_f) + downconversion[unit]
            else:
                elmer_mem_per_f = int(original_mem_int * multiplier)
                elmer_mem_per_f = str(elmer_mem_per_f) + unit
            return elmer_mem_per_f

        def _divup(a, b):
            return -(a // -b)

        sbatch_parameters = workflow["sbatch_parameters"]

        parallelization_level = workflow["_parallelization_level"]
        n_simulations = workflow["_n_simulations"]
        _n_workers = int(sbatch_parameters.pop("n_workers", 1))

        if parallelization_level == "full_simulation":
            n_workers_full = _n_workers
            n_workers_elmer_only = 1
            n_simulations_gmsh = n_simulations
        elif parallelization_level == "elmer":
            n_workers_full = 1
            n_workers_elmer_only = _n_workers
            n_simulations_gmsh = 1
        elif parallelization_level == "none":
            n_workers_full = 1
            n_workers_elmer_only = 1
            n_simulations_gmsh = 1
        else:
            logging.warning(f"Unknown parallelization level {parallelization_level}")

        if sbatch_parameters.get("--account", "project_0") == "project_0":
            sbatch_parameters["--account"] = KQC_REMOTE_ACCOUNT

        common_keys = [k for k in sbatch_parameters.keys() if k.startswith("--")]
        sbatch_settings_elmer = {k: sbatch_parameters.pop(k) for k in common_keys}

        sbatch_settings_meshes = sbatch_settings_elmer.copy()

        max_cpus_per_node = int(sbatch_parameters.pop("max_threads_per_node", 40))

        elmer_tasks_per_worker = int(sbatch_parameters.pop("elmer_n_processes", 10))
        elmer_cpus_per_task = int(sbatch_parameters.pop("elmer_n_threads", 1))
        if elmer_cpus_per_task > 1 and elmer_tasks_per_worker > 1:
            logging.warning(
                "Using both process and thread level parallelization" " with Elmer might result in poor performance"
            )
        elmer_cpus_per_worker = elmer_tasks_per_worker * elmer_cpus_per_task

        elmer_mem_per_worker = sbatch_parameters.pop("elmer_mem", "64G")

        if elmer_cpus_per_worker > max_cpus_per_node:
            elmer_nodes_per_worker = _divup(elmer_cpus_per_worker, max_cpus_per_node)
            elmer_total_nodes = elmer_nodes_per_worker * _n_workers
            elmer_tasks_per_node = min(elmer_tasks_per_worker, max_cpus_per_node)
            elmer_mem_per_node = _multiply_mem(elmer_mem_per_worker, _n_workers / elmer_total_nodes)
        else:
            elmer_nodes_per_worker = 1
            elmer_max_workers_per_node = max_cpus_per_node // elmer_cpus_per_worker
            elmer_total_nodes = _divup(_n_workers, elmer_max_workers_per_node)
            elmer_workers_per_node = min(_n_workers, elmer_max_workers_per_node)
            elmer_tasks_per_node = elmer_workers_per_node * elmer_tasks_per_worker
            elmer_mem_per_node = _multiply_mem(elmer_mem_per_worker, elmer_workers_per_node)

        gmsh_cpus_per_worker = int(sbatch_parameters.pop("gmsh_n_threads", 10))
        if gmsh_cpus_per_worker > max_cpus_per_node:
            raise RuntimeError(
                "Requested more gmsh threads per worker {gmsh_cpus_per_worker}"
                " than the limit per node {max_cpus_per_node}"
            )
        gmsh_max_workers_per_node = max_cpus_per_node // gmsh_cpus_per_worker
        gmsh_n_nodes = _divup(n_workers_full, gmsh_max_workers_per_node)
        gmsh_mem_per_worker = sbatch_parameters.pop("gmsh_mem", "64G")

        gmsh_workers_per_node = min(n_workers_full, gmsh_max_workers_per_node)
        gmsh_cpus_per_node = gmsh_cpus_per_worker * gmsh_workers_per_node
        gmsh_mem_per_node = _multiply_mem(gmsh_mem_per_worker, gmsh_workers_per_node)

        sbatch_settings_elmer["--time"] = _multiply_time(
            sbatch_parameters.pop("elmer_time", "00:10:00"), _divup(n_simulations, _n_workers)
        )
        sbatch_settings_elmer["--nodes"] = elmer_total_nodes
        sbatch_settings_elmer["--ntasks-per-node"] = elmer_tasks_per_node
        sbatch_settings_elmer["--cpus-per-task"] = elmer_cpus_per_task
        sbatch_settings_elmer["--mem"] = elmer_mem_per_node

        sbatch_settings_meshes["--time"] = _multiply_time(
            sbatch_parameters.pop("gmsh_time", "00:10:00"), _divup(n_simulations_gmsh, n_workers_full)
        )
        sbatch_settings_meshes["--nodes"] = gmsh_n_nodes
        sbatch_settings_meshes["--ntasks-per-node"] = 1
        sbatch_settings_meshes["--cpus-per-task"] = gmsh_cpus_per_node
        sbatch_settings_meshes["--mem"] = gmsh_mem_per_node

        if len(sbatch_parameters) > 0:
            logging.warning("Unused sbatch parameters: ")
            for k, v in sbatch_parameters.items():
                logging.warning(f"{k} : {v}")

        main_script_filename_meshes = str(path.joinpath(file_prefix + "_meshes.sh"))
        with open(main_script_filename_meshes, "w") as main_file:
            main_file.write("#!/bin/bash\n")

            for s_key, s_value in sbatch_settings_meshes.items():
                main_file.write(f"#SBATCH {s_key}={s_value}\n")

            main_file.write("\n")
            main_file.write("# set the number of threads based on --cpus-per-task\n")
            main_file.write("export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n")
            main_file.write("\n")

            for i, json_filename in enumerate(json_filenames):
                with open(json_filename) as f:
                    json_data = json.load(f)
                    simulation_name = json_data["name"]

                script_filename_meshes = str(path.joinpath(simulation_name + "_meshes.sh"))
                with open(script_filename_meshes, "w") as file:
                    file.write("#!/bin/bash\n")
                    file.write("set -e\n")
                    file.write('echo "Simulation {}/{} Gmsh"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        'srun -N 1 -n 1 -c {3} {2} "{0}" "{1}" --only-gmsh -q 2>&1 >> '
                        '"log_files/{4}.Gmsh.log"\n'.format(
                            execution_script,
                            Path(json_filename).relative_to(path),
                            python_executable,
                            str(gmsh_cpus_per_worker),
                            simulation_name,
                        )
                    )
                    file.write('echo "Simulation {}/{} ElmerGrid"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        'srun -N 1 -n 1 -c {2} ElmerGrid 14 2 "{0}.msh" 2>&1 >> '
                        '"log_files/{0}.ElmerGrid.log"\n'.format(
                            simulation_name,
                            Path(json_filename).relative_to(path),
                            str(gmsh_cpus_per_worker),
                        )
                    )
                    if int(elmer_tasks_per_worker) > 1:
                        file.write(
                            'srun -N 1 -n 1 -c {3} ElmerGrid 2 2 "{0}" 2>&1 -metis {1} 4'
                            ' --partdual --removeunused >> "log_files/{0}.ElmerGrid_part.log"\n'.format(
                                simulation_name,
                                str(elmer_tasks_per_worker),
                                Path(json_filename).relative_to(path),
                                str(gmsh_cpus_per_worker),
                            )
                        )
                    file.write('echo "Simulation {}/{} Write Elmer sif files"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        'srun -N 1 -n 1 -c {3} {2} "{0}" "{1}" --only-elmer-sifs '
                        '2>&1 >> "log_files/{4}.Elmer_sifs.log"\n'.format(
                            execution_script,
                            Path(json_filename).relative_to(path),
                            python_executable,
                            str(gmsh_cpus_per_worker),
                            simulation_name,
                        )
                    )
                os.chmod(script_filename_meshes, os.stat(script_filename_meshes).st_mode | stat.S_IEXEC)

                main_file.write('echo "Submitting gmsh and ElmerGrid part {}/{}"\n'.format(i + 1, len(json_filenames)))
                main_file.write('echo "--------------------------------------------"\n')
                main_file.write('source "{}" &\n'.format(Path(script_filename_meshes).relative_to(path)))
                if (i + 1) % n_workers_full == 0:
                    main_file.write("wait\n")

        # change permission
        os.chmod(main_script_filename_meshes, os.stat(main_script_filename_meshes).st_mode | stat.S_IEXEC)

        with open(main_script_filename, "w") as main_file:
            main_file.write("#!/bin/bash\n")

            for s_key, s_value in sbatch_settings_elmer.items():
                main_file.write(f"#SBATCH {s_key}={s_value}\n")

            main_file.write("\n")
            main_file.write("# set the number of threads based on --cpus-per-task\n")
            main_file.write("export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n")
            main_file.write("\n")
            for i, json_filename in enumerate(json_filenames):
                with open(json_filename) as f:
                    json_data = json.load(f)
                    simulation_name = json_data["name"]
                    sif_names = json_data["sif_names"]

                script_filename = str(path.joinpath(simulation_name + ".sh"))
                with open(script_filename, "w") as file:
                    file.write("#!/bin/bash\n")
                    file.write("set -e\n")
                    file.write('echo "Simulation {}/{} Elmer"\n'.format(i + 1, len(json_filenames)))

                    n_sifs = len(sif_names)
                    sifs_split = [
                        sif_names[i : min(i + n_workers_elmer_only, n_sifs)]
                        for i in range(0, n_sifs, n_workers_elmer_only)
                    ]
                    for sif_list in sifs_split:
                        for sif in sif_list:
                            file.write(
                                "srun --cpu-bind none --exact --mem={4} -N {5} -n {0} -c {3} "
                                'ElmerSolver_mpi "{1}/{2}.sif" 2>&1 >> "log_files/{2}.Elmer.log" & \n'.format(
                                    elmer_tasks_per_worker,
                                    simulation_name,
                                    sif,
                                    elmer_cpus_per_task,
                                    elmer_mem_per_worker,
                                    elmer_nodes_per_worker,
                                )
                            )
                        file.write("wait\n")

                    file.write('echo "Simulation {}/{} write results json"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        'srun -N 1 -n 1 -c {3} {2} "{0}" "{1}" --write-project-results 2>&1 >> '
                        "log_files/{4}.write_project_results.log\n".format(
                            execution_script,
                            Path(json_filename).relative_to(path),
                            python_executable,
                            1,
                            simulation_name,
                        )
                    )

                    file.write('echo "Simulation {}/{} write versions json"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        'srun -N 1 -n 1 -c {3} {2} "{0}" "{1}" --write-versions-file 2>&1 >> '
                        "log_files/{4}.write_versions_file.log\n".format(
                            execution_script,
                            Path(json_filename).relative_to(path),
                            python_executable,
                            1,
                            simulation_name,
                        )
                    )
                os.chmod(script_filename, os.stat(script_filename).st_mode | stat.S_IEXEC)

                main_file.write('echo "Submitting ElmerSolver part{}/{}"\n'.format(i + 1, len(json_filenames)))
                main_file.write('echo "--------------------------------------------"\n')
                main_file.write('source "{}" &\n'.format(Path(script_filename).relative_to(path)))
                if (i + 1) % n_workers_full == 0:
                    main_file.write("wait\n")

            main_file.write(get_post_process_command_lines(post_process, path, json_filenames))

    else:
        n_workers = workflow.get("n_workers", 1)
        parallelization_level = workflow["_parallelization_level"]
        parallelize_workload = parallelization_level == "full_simulation" and n_workers > 1

        with open(main_script_filename, "w") as main_file:
            main_file.write("#!/bin/bash\n")

            if parallelize_workload:
                main_file.write("export OMP_NUM_THREADS={}\n".format(workflow["elmer_n_threads"]))
                main_file.write("{} scripts/simple_workload_manager.py {}".format(python_executable, n_workers))

            for i, json_filename in enumerate(json_filenames):
                with open(json_filename) as f:
                    json_data = json.load(f)
                    simulation_name = json_data["name"]

                script_filename = str(path.joinpath(simulation_name + ".sh"))
                with open(script_filename, "w") as file:
                    file.write("#!/bin/bash\n")
                    file.write("set -e\n")
                    file.write('echo "Simulation {}/{} Gmsh"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        '{2} "{0}" "{1}" --only-gmsh 2>&1 >> "log_files/{3}.Gmsh.log"\n'.format(
                            execution_script, Path(json_filename).relative_to(path), python_executable, simulation_name
                        )
                    )
                    file.write('echo "Simulation {}/{} ElmerGrid"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        '{2} "{0}" "{1}" --only-elmergrid 2>&1 >> "log_files/{3}.ElmerGrid.log"\n'.format(
                            execution_script, Path(json_filename).relative_to(path), python_executable, simulation_name
                        )
                    )
                    file.write('echo "Simulation {}/{} Write Elmer sif files"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        '{2} "{0}" "{1}" --only-elmer-sifs 2>&1 >> "log_files/{3}.Elmer_sifs.log"\n'.format(
                            execution_script, Path(json_filename).relative_to(path), python_executable, simulation_name
                        )
                    )
                    file.write('echo "Simulation {}/{} Elmer"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        '{2} "{0}" "{1}" --only-elmer\n'.format(
                            execution_script, Path(json_filename).relative_to(path), python_executable
                        )
                    )
                    file.write('echo "Simulation {}/{} Paraview"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        '{2} "{0}" "{1}" --only-paraview\n'.format(
                            execution_script, Path(json_filename).relative_to(path), python_executable
                        )
                    )

                    file.write('echo "Simulation {}/{} write results json"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        '{2} "{0}" "{1}" --write-project-results 2>&1 >> '
                        '"log_files/{3}_write_project_results.log"\n'.format(
                            execution_script, Path(json_filename).relative_to(path), python_executable, simulation_name
                        )
                    )

                    file.write('echo "Simulation {}/{} write versions file"\n'.format(i + 1, len(json_filenames)))
                    file.write(
                        '{2} "{0}" "{1}" --write-versions-file\n'.format(
                            execution_script, Path(json_filename).relative_to(path), python_executable
                        )
                    )
                # change permission
                os.chmod(script_filename, os.stat(script_filename).st_mode | stat.S_IEXEC)
                if parallelize_workload:
                    main_file.write(' "./{}"'.format(Path(script_filename).relative_to(path)))
                else:
                    main_file.write(
                        'echo "Submitting the main script of simulation {}/{}"\n'.format(i + 1, len(json_filenames))
                    )
                    main_file.write('echo "--------------------------------------------"\n')
                    main_file.write('"./{}"\n'.format(Path(script_filename).relative_to(path)))

            main_file.write(get_post_process_command_lines(post_process, path, json_filenames))

    # change permission
    os.chmod(main_script_filename, os.stat(main_script_filename).st_mode | stat.S_IEXEC)

    return main_script_filename


def export_elmer(
    simulations: Sequence[
        Union[
            Simulation,
            Tuple[Simulation, ElmerSolution],
            CrossSectionSimulation,
            Tuple[CrossSectionSimulation, ElmerSolution],
        ]
    ],
    path: Path,
    script_folder: str = "scripts",
    file_prefix: str = "simulation",
    script_file: str = "run.py",
    workflow: Optional[Dict] = None,
    skip_errors: bool = False,
    post_process: Optional[Union[PostProcess, Sequence[PostProcess]]] = None,
    **solution_params,
) -> Path:
    """
    Exports an elmer simulation model to the simulation path.

    Args:
        simulations: List of Simulation objects or tuples containing Simulation and Solution objects.
        path: Location where to output the simulation model
        script_folder: Path to the Elmer-scripts folder.
        file_prefix: File prefix of the script file to be created.
        script_file: Name of the script file to run.
        workflow: Parameters for simulation workflow
        skip_errors: Skip simulations that cause errors. (Default: False)

            .. warning::

               **Use this carefully**, some of your simulations might not make sense physically and
               you might end up wasting time on bad simulations.
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations
        solution_params: ElmerSolution parameters if simulations is a list of Simulation objects.

    Returns:
        Path to exported script file.
    """
    export_with_tuples = isinstance(simulations[0], Sequence)
    if export_with_tuples:
        common_solution = None
    else:
        common_solution = ElmerSolution(**solution_params)

    _update_elmer_workflow(simulations, common_solution, workflow)

    write_commit_reference_file(path)
    copy_content_into_directory(ELMER_SCRIPT_PATHS, path, script_folder)
    json_filenames = []

    for sim_sol in simulations:
        simulation, solution = sim_sol if export_with_tuples else (sim_sol, common_solution)

        try:
            json_filenames.append(export_elmer_json(simulation, solution, path, workflow))
        except (IndexError, ValueError, Exception) as e:  # pylint: disable=broad-except
            if skip_errors:
                logging.warning(
                    f"Simulation {simulation.name} skipped due to {e.args}. "
                    "Some of your other simulations might not make sense geometrically. "
                    "Disable `skip_errors` to see the full traceback."
                )
            else:
                raise UserWarning(
                    "Generating simulation failed. You can discard the errors using `skip_errors` in `export_elmer`. "
                    "Moreover, `skip_errors` enables visual inspection of failed and successful simulation "
                    "geometry files."
                ) from e

    return export_elmer_script(
        json_filenames,
        path,
        workflow,
        file_prefix=file_prefix,
        execution_script=Path(script_folder).joinpath(script_file),
        post_process=post_process,
    )


def _update_elmer_workflow(simulations, common_solution, workflow):
    """
    Modify workflow based on number of simulations and available computing resources

    Args:
        simulations: List of Simulation objects or tuples containing Simulation and Solution objects.
        common_solution: Solution object if not contained in `simulations`
        workflow: workflow to be updated
    """
    if workflow is not None:
        parallelization_level = "none"
        n_worker_lim = 1
        num_sims = len(simulations)

        if num_sims == 1:
            if common_solution is None:
                num_freqs = len(simulations[0][1].frequency)
            else:
                num_freqs = len(common_solution.frequency)

            if num_freqs > 1:
                parallelization_level = "elmer"
                n_worker_lim = num_freqs
        else:  # num_sims > 1
            # TODO enable Elmer level parallelism
            n_worker_lim = num_sims
            parallelization_level = "full_simulation"

        if "sbatch_parameters" in workflow:
            n_workers = workflow["sbatch_parameters"].get("n_workers", 1.0)
            workflow["sbatch_parameters"]["n_workers"] = min(int(n_workers), n_worker_lim)
            workflow.pop("elmer_n_processes", "")
            workflow.pop("elmer_n_threads", "")
            workflow.pop("n_workers", "")
            workflow.pop("gmsh_n_threads", "")
        else:

            n_workers = workflow.get("n_workers", 1)
            n_processes = workflow.get("elmer_n_processes", 1)
            n_threads = workflow.get("elmer_n_threads", 1)

            if n_processes > 1 and n_threads > 1:
                logging.warning(
                    "Using both process and thread level parallelization" " with Elmer might result in poor performance"
                )

            # for the moment avoid psutil.cpu_count(logical=False)
            max_cpus = int(os.cpu_count() / 2 + 0.5)
            workflow["local_machine_cpu_count"] = max_cpus

            if n_workers == -1:
                n_processes = 1 if n_processes == -1 else n_processes
                n_threads = 1 if n_threads == -1 else n_threads
                n_workers = max(max_cpus // (n_threads * n_processes), 1)
                n_workers = min(n_workers, n_worker_lim)
            elif n_processes == -1:
                n_workers = min(n_workers, n_worker_lim)
                n_threads = 1 if n_threads == -1 else n_threads
                n_processes = max(max_cpus // (n_threads * n_workers), 1)
            elif n_threads == -1:
                n_workers = min(n_workers, n_worker_lim)
                n_threads = max(max_cpus // (n_processes * n_workers), 1)

            requested_cpus = n_workers * n_processes * n_threads
            if requested_cpus > max_cpus:
                logging.warning(f"Requested more CPUs ({requested_cpus}) than available ({max_cpus})")

            workflow["n_workers"] = n_workers
            workflow["elmer_n_processes"] = n_processes
            workflow["elmer_n_threads"] = n_threads
            workflow["_parallelization_level"] = parallelization_level
            workflow["_n_simulations"] = n_worker_lim

            gmsh_n_threads = workflow.get("gmsh_n_threads", 1)
            if gmsh_n_threads == -1:
                if parallelization_level == "full_simulation":
                    workflow["gmsh_n_threads"] = max(max_cpus // n_workers, 1)
                else:
                    workflow["gmsh_n_threads"] = max_cpus
    else:
        workflow = {}

    parser = argparse.ArgumentParser()
    parser.add_argument("-q", "--quiet", action="store_true")
    args, _ = parser.parse_known_args()

    if args.quiet:
        workflow.update(
            {
                "run_gmsh_gui": False,  # For GMSH: if true, the mesh is shown after it is done
                # (for large meshes this can take a long time)
                "run_paraview": False,  # this is visual view of the results
            }
        )



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import logging
from itertools import product
from pathlib import Path
from shutil import copytree

from kqcircuits.simulations.export.util import export_layers


def get_combined_parameters(simulation, solution):
    """Return parameters of Simulation and Solution in a combined dictionary.
    In case of common keys, 'solution.' prefix is added to Solution parameter key.
    """
    sim_dict = simulation.get_parameters()
    sol_dict = solution.__dict__
    return {
        **{k: v for k, v in sim_dict.items() if k != "name"},
        **{f"solution.{k}" if k in sim_dict else k: v for k, v in sol_dict.items() if k != "name"},
    }


def copy_content_into_directory(source_paths: list, path: Path, folder):
    """Create a folder and copy the contents of the source folders into it

    Arguments:
        source_paths: list of source directories from which to copy content
        path: path where the new folder will be created
        folder: name of the new folder
    """
    if path.exists() and path.is_dir():
        for source_path in source_paths:
            copytree(str(source_path), str(path.joinpath(folder)), dirs_exist_ok=True)


def get_post_process_command_lines(post_process, path, json_filenames):
    """Return post process command line calls as string. Can be used in construction of .bat or .sh script files.

    Args:
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations
        path: simulation folder path
        json_filenames: list of paths to simulation json files

    Returns:
        Command lines as string
    """
    if post_process is None:
        return ""

    commands = "echo Post-process\n"
    if isinstance(post_process, list):
        for pp in post_process:
            commands += pp.get_command_line(path, json_filenames)
    else:
        commands += post_process.get_command_line(path, json_filenames)
    return commands


def export_simulation_oas(simulations, path: Path, file_prefix="simulation"):
    """
    Write single OASIS file containing all simulations in list.
    """
    simulations = [simulation[0] if isinstance(simulation, tuple) else simulation for simulation in simulations]
    unique_layouts = {simulation.layout for simulation in simulations}
    if len(unique_layouts) != 1:
        raise ValueError("Cannot write batch OASIS file since not all simulations are on the same layout.")

    cells = [simulation.cell for simulation in simulations]
    oas_filename = str(path.joinpath(file_prefix + ".oas"))
    export_layers(oas_filename, simulations[0].layout, cells, output_format="OASIS", layers=None)
    return oas_filename


def sweep_simulation(layout, sim_class, sim_parameters, sweeps):
    """Create simulation sweep by varying one parameter at time. Return list of simulations."""
    simulations = []
    lengths = [len(l) for l in sweeps.values()]
    logging.info(f'Added simulations: {" + ".join([str(l) for l in lengths])} = {sum(lengths)}')
    for param in sweeps:
        for value in sweeps[param]:
            parameters = {
                **sim_parameters,
                param: value,
                "name": "{}_{}_{}".format(sim_parameters.get("name", ""), param, str(value)),
            }
            simulations.append(sim_class(**parameters) if layout is None else sim_class(layout, **parameters))
    return simulations


def cross_sweep_simulation(layout, sim_class, sim_parameters, sweeps):
    """Create simulation sweep by cross varying all parameters. Return list of simulations."""
    simulations = []
    keys = list(sweeps)
    sets = [list(prod) for prod in product(*sweeps.values())]
    logging.info(f'Added simulations: {" * ".join([str(len(l)) for l in sweeps.values()])} = {len(sets)}')
    for values in sets:
        parameters = {**sim_parameters}
        for i, key in enumerate(keys):
            parameters[key] = values[i]
        parameters["name"] = sim_parameters.get("name", "") + "_" + "_".join([str(value) for value in values])
        simulations.append(sim_class(**parameters) if layout is None else sim_class(layout, **parameters))
    return simulations



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import json
from pathlib import Path


class PostProcess:
    """Base class for adding post-processing scripts into the simulation batch"""

    def __init__(
        self,
        script,
        command="python",
        arguments="",
        folder="scripts",
        repeat_for_each=False,
        data_file_prefix=None,
        **data,
    ):
        """
        Args:
            script: name of the post-processing script
            command: command to run the post-processing script
            arguments: command line arguments for the post-processing script given as string
            folder: path where to look for the post-processing script file
            repeat_for_each: whether to repeat the post-processing script for every simulation. The simulation json file
                name becomes the first command line argument.
            data_file_prefix: prefix of the saved data file if data is given
            data: additional data to be saved into a file. The data file name becomes the last command line argument.
        """
        self.script = script
        self.command = command
        self.arguments = arguments
        self.folder = folder
        self.repeat_for_each = repeat_for_each
        self.data_file_prefix = data_file_prefix
        self.data = data

    def get_command_line(self, path, json_filenames):
        """Saves the data into file if needed and returns the command line to execute the post-processing script.

        Args:
            path: simulation folder path
            json_filenames: list of paths to simulation json files
        """
        str_args = self.arguments
        if self.data:
            file = str(Path(self.script).stem if self.data_file_prefix is None else self.data_file_prefix) + ".json"
            str_args += ' "' + file + '"'
            with open(path.joinpath(file), "w") as f:
                json.dump(self.data, f, indent=4)

        # Return the command line(s)
        str_cmd = f'{self.command} "{Path(self.folder).joinpath(self.script)}"'
        if self.repeat_for_each:
            lines = ""
            for json_filename in json_filenames:
                lines += f'{str_cmd} "{Path(json_filename).relative_to(path)}" {str_args}\n'
            return lines
        return f"{str_cmd} {str_args}\n"



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from math import sqrt
from kqcircuits.util.parameters import Param, pdt, add_parameters_from, pya
from kqcircuits.elements.element import Element

from kqcircuits.elements.waveguide_composite import WaveguideComposite
from kqcircuits.elements.waveguide_composite import Node
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter
from kqcircuits.elements.finger_capacitor_taper import FingerCapacitorTaper
from kqcircuits.util.refpoints import RefpointToEdgePort


@add_parameters_from(FingerCapacitorSquare, finger_gap_end=3)
@add_parameters_from(FingerCapacitorTaper, finger_width=10, finger_number=2, finger_gap=3)
class CapacitiveXCoupler(Element):
    """
    Capacitive coupler for testing FEM computations.
    """

    x_coupler_length = Param(pdt.TypeDouble, "Length of Capacitive X Coupler", 500)
    x_coupler_height = Param(pdt.TypeDouble, "Height of Capacitive X Coupler", 500)
    x_coupler_variant = Param(pdt.TypeString, "Coupler variant, either (+) or (x)", "+", choices=["+", "x"])
    remove_capacitors = Param(pdt.TypeBoolean, "Remove capacitors from the X Coupler", False)

    def build(self):
        length = self.x_coupler_length
        height = self.x_coupler_height
        if length > height:
            center = height
            input_length = (length - center) / 2.0
        else:
            center = length
            input_length = (height - center) / 2.0

        p0 = (0, 0)

        if self.x_coupler_variant == "x":
            if height > length:
                p11 = (-center / 2.0, -input_length - length / 2.0)
                p12 = (-length / 2.0, -center / 2.0)
                p13 = (-length / 4.0, -center / 4.0)
            else:
                p11 = (-input_length - center / 2.0, -height / 2.0)
                p12 = (-center / 2.0, -height / 2.0)
                p13 = (-center / 4.0, -height / 4.0)

            p21 = (-p11[0], p11[1])
            p22 = (-p12[0], p12[1])
            p23 = (-p13[0], p13[1])

            p31 = (-p11[0], -p11[1])
            p32 = (-p12[0], -p12[1])
            p33 = (-p13[0], -p13[1])

            p41 = (p11[0], -p11[1])
            p42 = (p12[0], -p12[1])
            p43 = (p13[0], -p13[1])

            p02 = ((self.a / 2 + self.b) / sqrt(2), (-self.a / 2 - self.b) / sqrt(2))
            p04 = ((-self.a / 2 - self.b) / sqrt(2), (self.a / 2 + self.b) / sqrt(2))
        elif self.x_coupler_variant == "+":
            p11 = (0.0, -height / 2.0)
            p12 = (0.0, -height / 3.0)
            p13 = (0.0, -height / 4.0)

            p21 = (length / 2.0, 0.0)
            p22 = (length / 3.0, 0.0)
            p23 = (length / 4.0, 0.0)

            p31 = (0.0, -p11[1])
            p32 = (0.0, -p12[1])
            p33 = (0.0, -p13[1])

            p41 = (-p21[0], 0.0)
            p42 = (-p22[0], 0.0)
            p43 = (-p23[0], 0.0)

            p02 = (self.a / 2 + self.b, 0.0)
            p04 = (-self.a / 2 - self.b, 0.0)

        self.refpoints["p11"] = pya.DPoint(*p11)
        self.refpoints["p21"] = pya.DPoint(*p21)
        self.refpoints["p31"] = pya.DPoint(*p31)
        self.refpoints["p41"] = pya.DPoint(*p41)

        def _add_capacitor_node(p):
            return Node(
                p,
                FingerCapacitorSquare,
                finger_number=self.finger_number,
                finger_width=self.finger_width,
                finger_gap_end=self.finger_gap_end,
                finger_gap=self.finger_gap,
                a2=self.a,
                b2=self.b,
                a=self.a,
                b=self.b,
                ground_padding=self.b,
            )

        splitter_port_length = self.a / 2 + self.b

        nodes1 = [
            Node(p11),
            Node(p12),
            _add_capacitor_node(p13),
            Node(
                p0,
                WaveguideCoplanarSplitter,
                angles=[0, 180, 90, 270],
                lengths=4 * [splitter_port_length],
                a=self.a,
                b=self.b,
            ),
            _add_capacitor_node(p33),
            Node(p32),
            Node(p31),
        ]
        nodes2 = [Node(p41), Node(p42), _add_capacitor_node(p43), Node(p04)]
        nodes3 = [Node(p21), Node(p22), _add_capacitor_node(p23), Node(p02)]

        if self.remove_capacitors:
            nodes1.pop(4)
            nodes1.pop(2)
            nodes2.pop(2)
            nodes3.pop(2)

        if self.x_coupler_variant == "x" and length == height:
            nodes1.pop(1)
            nodes1.pop(-1)
            nodes2.pop(1)
            nodes3.pop(1)

        self.insert_cell(WaveguideComposite, nodes=nodes1)
        self.insert_cell(WaveguideComposite, nodes=nodes2)
        self.insert_cell(WaveguideComposite, nodes=nodes3)

    @classmethod
    def get_sim_ports(cls, simulation):  # pylint: disable=unused-argument
        return [
            RefpointToEdgePort("p11"),
            RefpointToEdgePort("p21"),
            RefpointToEdgePort("p31"),
            RefpointToEdgePort("p41"),
        ]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.chip import Chip
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(Chip, frames_enabled=[0, 1])
class CrossingTwoface(Chip):
    """The PCell declaration for a CrossingTwoFace chip.

    On the left side of the chip there is a straight vertical waveguide bottom face and a meandering waveguide crossing
    multiple times on the top face. There are transmission lines at different faces at the crossings. On the right side
    there is likewise a straight and a meandering waveguide, but they do not cross at any point.
    """

    crossings = Param(pdt.TypeInt, "Number of double crossings", 3, docstring="Number of pairs of flip-chip crossings")
    crossing_length = Param(
        pdt.TypeDouble,
        "Crossing waveguide length",
        400,
        unit="μm",
        docstring="Length of the crossing on the top face (μm)",
    )
    cross_talk_distance = Param(
        pdt.TypeDouble,
        "Transmission line distance from meander",
        300,
        unit="μm",
        docstring="Distance between the right straight transmission line and meander on the right (μm)",
    )
    meander_face = Param(pdt.TypeString, "Meander face on right side", "single", choices=["Single", "Two Face"])

    def build(self):
        launchers = self.produce_launchers(
            "SMA8",
            launcher_assignments={
                1: "PL-1-IN",
                2: "PL-2-IN",
                3: "PL-4-IN",
                4: "PL-4-OUT",
                5: "PL-2-OUT",
                6: "PL-1-OUT",
                7: "PL-3-OUT",
                8: "PL-3-IN",
            },
        )
        self._produce_transmission_lines(launchers)

    def _produce_transmission_lines(self, launchers):
        distance = 700
        right_tr_x = 5000 + distance
        left_tr_x = 5000 - distance
        face1_box = self.get_box(1)

        # Left transmission line
        nodes = [
            Node(self.refpoints["PL-1-IN_port"]),
            Node(self.refpoints["PL-1-IN_port_corner"] + pya.DPoint(0, -2 * self.r)),
            Node((left_tr_x, self.refpoints["PL-1-IN_port_corner"].y - 2 * self.r)),
            Node((left_tr_x, face1_box.p2.y), a=self.a_capped, b=self.b_capped),
            Node((left_tr_x, face1_box.p1.y + 100), a=self.a, b=self.b),
            Node((left_tr_x, self.refpoints["PL-1-OUT_port_corner"].y + 2 * self.r)),
            Node(self.refpoints["PL-1-OUT_port_corner"] + pya.DPoint(0, 2 * self.r)),
            Node(self.refpoints["PL-1-OUT_port"]),
        ]

        self.insert_cell(WaveguideComposite, nodes=nodes)

        # Right transmission line
        nodes = [
            Node(self.refpoints["PL-2-IN_port"]),
            Node(self.refpoints["PL-2-IN_port_corner"] + pya.DPoint(0, -2 * self.r)),
            Node((right_tr_x, self.refpoints["PL-2-IN_port_corner"].y - 2 * self.r)),
            Node((right_tr_x, face1_box.p2.y), a=self.a_capped, b=self.b_capped),
            Node((right_tr_x, face1_box.p1.y + 100), a=self.a, b=self.b),
            Node((right_tr_x, self.refpoints["PL-2-OUT_port_corner"].y + 2 * self.r)),
            Node(self.refpoints["PL-2-OUT_port_corner"] + pya.DPoint(0, 2 * self.r)),
            Node(self.refpoints["PL-2-OUT_port"]),
        ]

        self.insert_cell(WaveguideComposite, nodes=nodes)

        # Crossing transmission line
        nodes = [
            Node(self.refpoints["PL-3-IN_port"]),
            Node((face1_box.p1.x, self.refpoints["PL-3-IN_port"].y), a=self.a_capped, b=self.b_capped),
        ]
        ref_x = left_tr_x
        ref_x_1 = ref_x - self.crossing_length / 2.0
        ref_x_2 = ref_x + self.crossing_length / 2.0

        last_y = launchers["PL-3-IN"][0].y
        crossings = self.crossings  # must be even
        step = (launchers["PL-3-IN"][0].y - launchers["PL-3-OUT"][0].y) / (crossings - 0.5) / 2
        wiggle = self.crossing_length / 2.0 + 250

        for i in range(crossings):
            nodes.append(Node((ref_x - wiggle, last_y)))
            nodes.append(Node((ref_x_1, last_y), face_id=self.face_ids[1]))
            nodes.append(Node((ref_x_2, last_y), face_id=self.face_ids[0]))
            nodes.append(Node((ref_x + wiggle, last_y)))
            last_y -= step
            nodes.append(Node((ref_x + wiggle, last_y)))
            nodes.append(Node((ref_x_2, last_y), face_id=self.face_ids[1]))
            nodes.append(Node((ref_x_1, last_y), face_id=self.face_ids[0]))
            nodes.append(Node((ref_x - wiggle, last_y)))
            last_y -= step
        nodes.append(Node((face1_box.p1.x + 100, self.refpoints["PL-3-OUT_port"].y), a=self.a, b=self.b))
        nodes.append(Node(self.refpoints["PL-3-OUT_port_corner"]))
        nodes.append(Node(self.refpoints["PL-3-OUT_port"]))
        self.insert_cell(WaveguideComposite, nodes=nodes)

        # cross_talk
        ref_x = right_tr_x + self.cross_talk_distance + wiggle
        last_y = self.refpoints["PL-4-IN_port"].y
        nodes = [
            Node(self.refpoints["PL-4-IN_port"]),
            Node(self.refpoints["PL-4-IN_port_corner"]),
            Node((face1_box.p2.x, self.refpoints["PL-4-IN_port"].y), a=self.a_capped, b=self.b_capped),
        ]
        for i in range(crossings):
            if i == 0 and self.meander_face == "Two Face":
                nodes.append(Node((ref_x + wiggle, last_y), face_id=self.face_ids[1]))
            else:
                nodes.append(Node((ref_x + wiggle, last_y)))
            nodes.append(Node((ref_x - wiggle, last_y)))
            last_y -= step
            nodes.append(Node((ref_x - wiggle, last_y)))
            if i == range(crossings)[-1] and self.meander_face == "Two Face":
                nodes.append(Node((ref_x + wiggle, last_y), face_id=self.face_ids[0]))
            else:
                nodes.append(Node((ref_x + wiggle, last_y)))
            last_y -= step
        nodes.append(Node((face1_box.p2.x - 100, self.refpoints["PL-4-OUT_port"].y), a=self.a, b=self.b))
        nodes.append(Node(self.refpoints["PL-4-OUT_port_corner"]))
        nodes.append(Node(self.refpoints["PL-4-OUT_port"]))
        self.insert_cell(WaveguideComposite, nodes=nodes)



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import argparse
import logging
import sys
import subprocess
from os import listdir, chdir
from os.path import isfile, join
from pathlib import Path
from kqcircuits.simulations.export.export_and_run import export_and_run
from kqcircuits.simulations.export.export_singularity import export_singularity
from kqcircuits.simulations.export.remote_export_and_run import remote_export_and_run, remote_run_only
from kqcircuits.defaults import TMP_PATH, SCRIPTS_PATH, ROOT_PATH

logging.basicConfig(level=logging.WARN, stream=sys.stdout)


def argument_parser():
    parser = argparse.ArgumentParser(description="KQC console scripts")

    subparser = parser.add_subparsers(dest="command")

    simulate_parser = subparser.add_parser(
        "sim",
        help="KQC simulation \
            export and run script. Run and export with a single command!",
    )
    mask_parser = subparser.add_parser("mask", help="Build KQC Mask.")

    singularity_parser = subparser.add_parser(
        "singularity",
        help="Build and push \
                                              singularity image to remote host.",
    )

    simulate_parser.add_argument("export_script", type=str, help="Name of the export script")
    simulate_parser.add_argument(
        "--export-path-basename",
        type=str,
        default=None,
        help="The export folder will be `TMP_PATH / Path(export_path_basename)`, \
                                       if not given, then it will be `TMP_PATH / Path(args.export_script).stem`",
    )
    simulate_parser.add_argument("-q", "--quiet", action="store_true", help="Quiet mode: No GUI")
    simulate_parser.add_argument(
        "-e", "--export-only", action="store_true", help="Do not run simulation, only export generated files."
    )

    simulate_parser.add_argument(
        "--remote", action="store_true", help='Export and run the simulations at remote host "user@host"'
    )
    simulate_parser.add_argument(
        "--remote-run-only", action="store_true", help='Run the simulation at remote host "user@host"'
    )
    simulate_parser.add_argument("--kqc-remote-tmp-path", type=str, help="Path to the used tmp directory on remote")

    simulate_parser.add_argument(
        "--detach", action="store_true", help="Detach the remote simulation from terminal, not waiting for it to finish"
    )
    simulate_parser.add_argument(
        "--poll-interval", type=int, help="Interval for polling the job state of remote simulation in seconds"
    )

    mask_parser.add_argument("mask_script", type=str, help="Name of the mask script")
    mask_parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Debug mode. Use a single process and " "print logs to standard output too.",
    )
    mask_parser.add_argument(
        "-m",
        "--mock_chips",
        action="store_true",
        help="Replace all chips with empty frames for " "faster testing of the mask layout",
    )
    mask_parser.add_argument("-s", "--skip_extras", action="store_true", help="Skip netlist and documentation export")
    mask_parser.add_argument("-c N", action="store_true", help="Limit the number of used CPUs to 'N'")

    singularity_parser.add_argument("--build", action="store_true", help="build singularity image locally")
    singularity_parser.add_argument(
        "--push",
        type=str,
        help='Destination of the export "user@host:dir"\
                                    if left empty the defaults from .remote_profile.txt will be used',
    )
    singularity_parser.add_argument(
        "--singularity-remote-path", type=str, help="Installation path for the singularity image on remote"
    )
    return parser, subparser


def run_kqc(args, args_for_script):
    if args.command == "sim":
        if args.export_script == "ls":
            simdir = Path(SCRIPTS_PATH / "simulations")
            for f in listdir(simdir):
                if isfile(join(simdir, f)):
                    print(f)
            return True

        if args.remote:
            remote_host = str(args.export_script)
            remote_export_and_run(
                remote_host,
                args.kqc_remote_tmp_path,
                args.detach,
                args.poll_interval,
                args.export_path_basename,
                args_for_script,
            )
            return True
        if args.remote_run_only:
            remote_host = str(args.export_script)
            remote_run_only(remote_host, args.kqc_remote_tmp_path, args.detach, args.poll_interval, args_for_script)
            return True

        script_file = Path(args.export_script)
        if args.export_path_basename is not None:
            export_path = TMP_PATH / args.export_path_basename
        else:
            export_path = TMP_PATH / script_file.stem

        if not script_file.is_file():
            script_file = Path(SCRIPTS_PATH / "simulations" / script_file)
            if not script_file.is_file():
                logging.error(f"Export script not found at {args.export_script} or {script_file}")
                return True
        export_and_run(script_file, export_path, args.quiet, args.export_only, args_for_script)
        return True
    elif args.command == "mask":
        if args.mask_script == "ls":
            maskdir = Path(SCRIPTS_PATH / "masks")
            for f in listdir(maskdir):
                if isfile(join(maskdir, f)):
                    print(f)
            return True
        script_file = Path(args.mask_script)
        if not script_file.is_file():
            script_file = Path(SCRIPTS_PATH / "masks" / args.mask_script)
            if not script_file.is_file():
                logging.error(f"Mask script not found at {args.mask_script} or {script_file}")
                return True
        if args.debug:
            subprocess.call([sys.executable, script_file, "-d"])
            return True
        else:  # Windows needs this for multiprocessing
            with open(script_file) as mask_file:
                exec(mask_file.read())  # pylint: disable=exec-used
            return True
    elif args.command == "singularity":
        if args.build:
            chdir(Path(ROOT_PATH / "singularity"))
            with open("install_software.sh", "r") as f:
                for line in f:
                    if "export MPI_VERSION=" in line:
                        mpi_v = line.partition("export MPI_VERSION=")[2].strip()
                        print(
                            (
                                f"Singularity will use MPI version {mpi_v}. "
                                "Make sure this corresponds to the MPI version on the target machine\n"
                                "MPI and other package versions used by singularity can be changed "
                                "in the beginning of the /singularity/install_software.sh script"
                            )
                        )
                        break
            subprocess.call("./singularity.sh", shell=True)
        elif args.push is not None:
            export_singularity(args.push, args.singularity_remote_path)
    else:
        return False
    return True


def run():
    """Main entry point for the KQC console command"""
    parser, _ = argument_parser()
    args, args_for_script = parser.parse_known_args()
    if not run_kqc(args, args_for_script):
        parser.print_usage()



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from dataclasses import dataclass, field
from typing import Union, List
from kqcircuits.simulations.export.solution import Solution


@dataclass
class ElmerSolution(Solution):
    """
    A Base class for Elmer Solution parameters

    Args:
        tool: Available: "capacitance", "wave_equation" and "cross-section"
        linear_system_method: Available: 'bicgstab', 'mg'. Currently only applies to Capacitance Solver
        p_element_order: polynomial order of p-elements
        percent_error: Stopping criterion in adaptive meshing.
        max_error_scale: Maximum element error, relative to percent_error, allowed in individual elements.
        max_outlier_fraction: Maximum fraction of outliers from the total number of elements
        maximum_passes: Maximum number of adaptive meshing iterations.
        minimum_passes: Minimum number of adaptive meshing iterations.
        is_axisymmetric: Simulate with Axi Symmetric coordinates along :math:`y\\Big|_{x=0}` (Default: False)
        mesh_levels: If set larger than 1 Elmer will make the mesh finer by dividing each element
                          into 2^(dim) elements mesh_levels times. Default 1.
        frequency: Units are in GHz. Give a list of frequencies if using interpolating sweep.
        frequency_batch: Number of frequencies calculated between each round of fitting in interpolating sweep
        sweep_type: Type of frequency sweep. Options "explicit" and "interpolating".
        max_delta_s: Convergence tolerance in interpolating sweep
        boundary_conditions: Parameters to determine boundary conditions
        integrate_energies: Calculate energy integrals over each object. Used in EPR simulations
        mesh_levels: If set larger than 1 Elmer will make the mesh finer by dividing each element
                          into 2^(dim) elements mesh_levels times. Default 1.
        mesh_size: Parameters to determine mesh element sizes
        solver_options: Can be used to set experimental solver options for Elmer wave-equation tool.
                        Supports the options `use_av` (bool), `london_penetration_depth` (float),
                        `conductivity` (float),`nested_iteration` (bool), `convergence_tolerance` (float),
                        `max_iterations` (int), `quadratic_approximation` (bool), `second_kind_basis` (bool)

    """

    tool: str = "capacitance"  # to be ClassVar
    linear_system_method: str = "bicgstab"
    p_element_order: int = 3
    percent_error: float = 0.005
    max_error_scale: float = 2.0
    max_outlier_fraction: float = 1e-3
    maximum_passes: int = 1
    minimum_passes: int = 1
    is_axisymmetric: bool = False
    mesh_levels: int = 1
    frequency: Union[float, List[float]] = 5
    frequency_batch: int = 3
    sweep_type: str = "explicit"
    max_delta_s: float = 0.01
    boundary_conditions: dict = field(default_factory=dict)
    integrate_energies: bool = False
    mesh_levels: int = 1
    mesh_size: dict = field(default_factory=dict)
    solver_options: dict = field(default_factory=dict)

    def __post_init__(self):
        """Cast frequency to list. Automatically called after init"""
        if isinstance(self.frequency, (float, int)):
            self.frequency = [self.frequency]
        elif not isinstance(self.frequency, list):
            self.frequency = list(self.frequency)

    def get_solution_data(self):
        """Return the solution data in dictionary form."""
        sol_dict = {**self.__dict__}
        sol_dict["solution_name"] = sol_dict.pop("name")
        return sol_dict



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from kqcircuits.pya_resolver import pya


def get_list_of_two(dims):
    """Returns list of two terms when 'dims' is given as a scalar or list."""
    if isinstance(dims, list):
        return [dims[0], dims[1]]
    return [dims, dims]


class PartitionRegion:
    """Class to enable partitioning of simulation geometry into sub-regions"""

    def __init__(
        self,
        name="part",
        region=None,
        z=None,
        face=None,
        vertical_dimensions=None,
        metal_edge_dimensions=None,
    ):
        """
        Args:
            name: Suffix of the partition layers. Must not end with a number.
            region: Area to which the partition region is limited. Can be given as pya.DBox, pya.DPolygon, or list of
                pya.DPolygons. Use None to cover full domain.
            z: Lower and upper bound for the partition region as scalar or list. Use None to cover full height.
            face: The face name to which the partition region is applied. If this is used, the vertical_dimensions
                and metal_edge_dimensions are applied.
            vertical_dimensions: Vertical dimensions of the partition region on face as scalar or list. The terms in the
                list correspond to expansion dimensions into directions of substrate and vacuum, respectively. Scalar
                means the substrate and vacuum expansions are equal. This is applied only if face is given.
            metal_edge_dimensions: Lateral dimensions to limit the partition region next to the metal edges. If given as
                list, the terms correspond to expansions into directions of gap and metal, respectively. If given as
                scalar, the gap and metal expansions are equal. Use None to disable the metal edge limitation. This is
                applied only if face is given.
        """
        if name[-1] in "0123456789":
            raise ValueError(f"PartitionRegion name must not end with a number, but {name} is given.")
        if name == "":
            raise ValueError("PartitionRegion name must not be an empty string.")
        self.name = name
        self.region = region
        self.z = z
        self.face = face
        self.vertical_dimensions = vertical_dimensions
        self.metal_edge_dimensions = metal_edge_dimensions

    def limit_box(self, bottom, top, box, dbu):
        """Limits the region and z-levels into simulation dimensions.

        Args:
            bottom: bottom of the simulation domain.
            top: top of the simulation domain.
            box: lateral dimensions of simulation domain as pya.DBox.
            dbu: layout database unit
        """
        # update self.z using bottom and top
        self.z = get_list_of_two(self.z)
        if self.z[0] is None or self.z[0] < bottom:
            self.z[0] = bottom
        if self.z[1] is None or top < self.z[1]:
            self.z[1] = top

        # update self.region using box
        box_region = pya.Region(box.to_itype(dbu))
        if self.region is None:
            self.region = box_region
        elif isinstance(self.region, list):
            merged_region = pya.Region()
            for r in self.region:
                merged_region += pya.Region(r.to_itype(dbu))
            self.region = merged_region & box_region
        else:
            self.region = pya.Region(self.region.to_itype(dbu)) & box_region

    def limit_face(self, z, sign, metal_region, etch_region, dbu):
        """Limits the region and z-levels on face. Function limit_box should be called once before this.

        Args:
            z: z-level of the face
            sign: 1 if substrate is below vacuum, -1 otherwise
            metal_region: metallization area as pya.Region
            etch_region: area where metal is etched away as pya.Region
            dbu: layout database unit
        """
        # Reset face to indicate that face limitation is applied
        self.face = None

        # update self.z using self.vertical_dimensions
        vd = get_list_of_two(self.vertical_dimensions)
        if vd[sign < 0] is not None and self.z[0] < z - vd[sign < 0]:
            self.z[0] = z - vd[sign < 0]
        if vd[sign > 0] is not None and z + vd[sign > 0] < self.z[1]:
            self.z[1] = z + vd[sign > 0]

        # update self.region using self.metal_edge_dimensions
        ed = get_list_of_two(self.metal_edge_dimensions)
        if ed[0] is not None:
            self.region &= metal_region.sized(ed[0] / dbu)
        if ed[1] is not None:
            self.region &= etch_region.sized(ed[1] / dbu)
